from __future__ import annotations

from pynyaa import Category, Nyaa


def test_properties(nyaa_client: Nyaa) -> None:
    assert nyaa_client.base_url == "https://nyaa.si/"


def test_nyaa_default(nyaa_client: Nyaa) -> None:
    nyaa = nyaa_client.get("https://nyaa.si/view/1755409")
    assert nyaa.title == "[smol] Shelter (2016) (BD 1080p HEVC FLAC) | Porter Robinson & Madeon - Shelter"
    assert nyaa.submitter.name == "smol"
    assert nyaa.submitter.is_trusted is False
    assert nyaa.submitter.is_banned is False


def test_nyaa_trusted(nyaa_client: Nyaa) -> None:
    nyaa = nyaa_client.get("https://nyaa.si/view/1544043")
    assert nyaa.title == "[MTBB] I Want to Eat Your Pancreas (BD 1080p) | Kimi no Suizou wo Tabetai"
    assert nyaa.submitter.is_trusted is True
    assert nyaa.is_remake is False
    assert nyaa.is_trusted is True
    assert nyaa.category == Category.ANIME_ENGLISH_TRANSLATED


def test_nyaa_trusted_and_remake(nyaa_client: Nyaa) -> None:
    nyaa = nyaa_client.get("https://nyaa.si/view/1694824")
    assert (
        nyaa.title
        == "[MiniMTBB] Sound! Euphonium the Movie: Our Promise: A Brand New Day (BD 1080p) | Hibike! Euphonium: Chikai no Finale"
    )
    assert nyaa.is_remake is True
    assert nyaa.is_trusted is False
    assert nyaa.submitter.is_trusted is True


def test_nyaa_anon(nyaa_client: Nyaa) -> None:
    nyaa = nyaa_client.get(1765655)
    assert (
        nyaa.title
        == "[Kinoworm] Ascendance of a Bookworm P01v01-P05v08 + Fanbook v01-v03 + Royal Academy Stories First Year v01 + Short Story Collection v01 (J-Novel Club) (Premium)"
    )
    assert nyaa.url.__str__() == "https://nyaa.si/view/1765655"
    assert nyaa.information == "https://www.goodreads.com/series/220639-ascendance-of-a-bookworm-light-novel"
    assert nyaa.submitter.name == "Anonymous"
    assert nyaa.category == Category.LITERATURE_ENGLISH_TRANSLATED


def test_nyaa_banned(nyaa_client: Nyaa) -> None:
    nyaa = nyaa_client.get("https://nyaa.si/view/1422797")
    assert nyaa.title == "[succ_] Tsugumomo [BDRip 1920x1080 x264 FLAC]"
    assert nyaa.submitter.is_trusted is False
    assert nyaa.submitter.is_banned is True
    assert len(nyaa.torrent.files) == 20


def test_nyaa_banned_and_trusted(nyaa_client: Nyaa) -> None:
    nyaa = nyaa_client.get("https://nyaa.si/view/884488")
    assert nyaa.title == "[FMA1394] Fullmetal Alchemist (2003) [Dual Audio] [US BD] (batch)"
    assert nyaa.submitter.is_trusted is True
    assert nyaa.submitter.is_banned is True
    assert len(nyaa.torrent.files) == 51


def test_nyaa_empty_info(nyaa_client: Nyaa) -> None:
    nyaa = nyaa_client.get("https://nyaa.si/view/5819")
    assert nyaa.information is None
    assert nyaa.description is not None


def test_nyaa_empty_desc(nyaa_client: Nyaa) -> None:
    nyaa = nyaa_client.get("https://nyaa.si/view/76777")
    assert nyaa.information is not None
    assert nyaa.description is None


def test_nyaa_empty_desc_info(nyaa_client: Nyaa) -> None:
    nyaa = nyaa_client.get("https://nyaa.si/view/1586776")
    assert nyaa.information is None
    assert nyaa.description is None


def test_nyaa_search_no_results(nyaa_client: Nyaa) -> None:
    results = nyaa_client.search("akldlaskdjsaljdksd")
    assert tuple(results) == tuple()


def test_search_single(nyaa_client: Nyaa) -> None:
    results = nyaa_client.search("smol shelter")

    for result in results:
        assert result.title == "[smol] Shelter (2016) (BD 1080p HEVC FLAC) | Porter Robinson & Madeon - Shelter"
        break



from __future__ import annotations

from typing import get_args

from pynyaa._types import CategoryID, CategoryName, CategoryValue, SortByLiteral


def test_category_name_literals() -> None:
    assert get_args(CategoryName) == (
        "ALL",
        "ANIME",
        "ANIME_MUSIC_VIDEO",
        "ANIME_ENGLISH_TRANSLATED",
        "ANIME_NON_ENGLISH_TRANSLATED",
        "ANIME_RAW",
        "AUDIO",
        "AUDIO_LOSSLESS",
        "AUDIO_LOSSY",
        "LITERATURE",
        "LITERATURE_ENGLISH_TRANSLATED",
        "LITERATURE_NON_ENGLISH_TRANSLATED",
        "LITERATURE_RAW",
        "LIVE_ACTION",
        "LIVE_ACTION_ENGLISH_TRANSLATED",
        "LIVE_ACTION_IDOL_PROMOTIONAL_VIDEO",
        "LIVE_ACTION_NON_ENGLISH_TRANSLATED",
        "LIVE_ACTION_RAW",
        "PICTURES",
        "PICTURES_GRAPHICS",
        "PICTURES_PHOTOS",
        "SOFTWARE",
        "SOFTWARE_APPLICATIONS",
        "SOFTWARE_GAMES",
    )


def test_category_value_literals() -> None:
    assert get_args(CategoryValue) == (
        "All",
        "Anime",
        "Anime - Anime Music Video",
        "Anime - English-translated",
        "Anime - Non-English-translated",
        "Anime - Raw",
        "Audio",
        "Audio - Lossless",
        "Audio - Lossy",
        "Literature",
        "Literature - English-translated",
        "Literature - Non-English-translated",
        "Literature - Raw",
        "Live Action",
        "Live Action - English-translated",
        "Live Action - Idol/Promotional Video",
        "Live Action - Non-English-translated",
        "Live Action - Raw",
        "Pictures",
        "Pictures - Graphics",
        "Pictures - Photos",
        "Software",
        "Software - Applications",
        "Software - Games",
    )


def test_category_id_literal() -> None:
    assert get_args(CategoryID) == (
        "0_0",
        "1_0",
        "1_1",
        "1_2",
        "1_3",
        "1_4",
        "2_0",
        "2_1",
        "2_2",
        "3_0",
        "3_1",
        "3_2",
        "3_3",
        "4_0",
        "4_1",
        "4_2",
        "4_3",
        "4_4",
        "5_0",
        "5_1",
        "5_2",
        "6_0",
        "6_1",
        "6_2",
    )


def test_category_literals_length() -> None:
    assert len(get_args(CategoryName)) == len(get_args(CategoryValue)) == len(get_args(CategoryID)) == 24


def test_sort_by_literals() -> None:
    assert get_args(SortByLiteral) == ("comments", "size", "id", "datetime", "seeders", "leechers", "downloads")



from __future__ import annotations

from os import fspath
from typing import Any

from pydantic import BaseModel, ConfigDict, HttpUrl, field_serializer, field_validator
from torf import Torrent

from pynyaa._enums import Category
from pynyaa._types import MagnetUrl, UTCDateTime


class ParentModel(BaseModel):
    """
    Parent model that stores global configuration
    All models ahead will inherit from this.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)


class Submitter(ParentModel):
    """
    Model representing the user who submitted the torrent.

    Features
    --------
    - Immutable
    - Hashable
    - Inherits from [`pydantic.BaseModel`][pydantic.BaseModel], so you get all of pydantic's fancy methods
    - Supports equality checking (based on the URL)

    Examples
    --------

    ```py
    >>> a = Submitter(name="John", url="https://nyaa.si/user/john", is_trusted=True, is_banned=False)
    >>> b = Submitter(name="John", url="https://nyaa.si/user/john", is_trusted=True, is_banned=False) # dupe
    >>> c = Submitter(name="Jane", url="https://nyaa.si/user/jane", is_trusted=False, is_banned=False)

    >>> print(a)
    John

    >>> a == b
    True

    >>> a == c
    False

    >>> set((a, b, c)) == {a, c} == {b, c} # dedupe
    True
    ```
    """

    name: str
    """Username of the submitter."""
    url: HttpUrl
    """Profile URL of the submitter."""
    is_trusted: bool
    """Indicates whether the user is trusted (green) or not."""
    is_banned: bool
    """Indicates whether the user is banned or not."""

    def __eq__(self, other: object) -> bool:
        """
        Implements equality method.
        Two Submitter(s) are equal if they both point to the same URL.
        """
        if not isinstance(other, Submitter):
            return NotImplemented
        return self.url == other.url

    def __hash__(self) -> int:
        """
        Makes Submitter hashable.
        """
        return hash(self.url)

    def __repr__(self) -> str:
        """
        This matches Submitter's repr with NyaaTorrentPage's for consistency.
        """
        return f"{self.__class__.__name__}(name='{self.name}', url='{self.url}', is_trusted={self.is_trusted}, is_banned={self.is_banned})"

    def __str__(self) -> str:
        """
        Stringify into something easily readable.
        """
        return self.name


class NyaaTorrentPage(ParentModel):
    """
    Model representing Nyaa's torrent page.

    Features
    --------
    - Immutable
    - Hashable
    - Inherits from [`pydantic.BaseModel`][pydantic.BaseModel], so you get all of pydantic's fancy methods
    - Supports equality checking (based on the URL)

    Examples
    --------

    ```py
    >>> from pynyaa import Nyaa
    >>> nyaa = Nyaa()
    >>> a = nyaa.get(1839783)
    >>> b = nyaa.get(1839783) # dupe
    >>> c = nyaa.get(1839609)

    >>> print(a)
    [SubsPlease] Hibike! Euphonium S3 - 13 (1080p) [230618C3].mkv

    >>> a == b
    True

    >>> a == c
    False

    >>> set((a, b, c)) == {a, c} == {b, c} # dedupe
    True
    ```
    """

    id: int
    """Nyaa ID of the torrent (`https://nyaa.si/view/{id}`)."""

    url: HttpUrl
    """URL to the Nyaa torrent page (`https://nyaa.si/view/123456`)."""

    title: str
    """Title of the torrent."""

    category: Category
    """Torrent category."""

    submitter: Submitter
    """User who submitted the torrent."""

    datetime: UTCDateTime
    """Date and time at which the torrent was submitted."""

    information: str | None
    """Information about the torrent."""

    seeders: int
    """Number of seeders."""

    leechers: int
    """Number of leechers."""

    completed: int
    """Number of completed downloads."""

    is_trusted: bool
    """Indicates whether the upload is trusted (green) or not."""

    is_remake: bool
    """
    Indicates whether the upload is a remake (red) or not.
    
    Note
    ----
    An upload can be both trusted and a remake, in which case,
    the remake takes priority, that is, `is_remake` will be `True`
    and `is_trusted` will be `False`.
    This is a current limitation that I don't know how to work around.
    """

    description: str | None
    """Torrent description."""

    torrent_file: HttpUrl
    """URL pointing to the `.torrent` file (`https://nyaa.si/download/123456.torrent`)"""

    magnet: MagnetUrl
    """
    Magnet link of the torrent. 
    
    Note
    ----
    The magnet link provided by Nyaa is different from the one 
    you'll get if you simply generated it from the `.torrent` file itself
    because Nyaa strips away all trackers except it's own 
    and the ones listed [here](https://github.com/nyaadevs/nyaa/blob/master/trackers.txt).
    """

    torrent: Torrent
    """
    A [`torf.Torrent`][torf.Torrent] object 
    representing the data stored in the `.torrent` file.
    """

    @field_serializer("torrent", when_used="json")
    def _serialize_torf_torrent(torrent: Torrent) -> dict[str, Any]:
        """
        JSON Serializer for torf.Torrent.
        """
        # Convert torf.Trackers object into a plain nested list
        trackers = []
        for tier in torrent.trackers:
            tierlist = []
            for url in tier:
                tierlist.append(url)
            trackers.append(tierlist)

        # Convert torf.Files object into a list of dictionaries
        files = []
        for file in torrent.files:
            files.append(dict(file=fspath(file), size=file.size))

        return dict(
            name=torrent.name,
            size=torrent.size,
            infohash=torrent.infohash,
            piece_size=torrent.piece_size,
            private=torrent.private,
            trackers=trackers,
            comment=torrent.comment,
            creation_date=torrent.creation_date,
            created_by=torrent.created_by,
            source=torrent.source,
            files=files,
        )

    @field_validator("information", "description")
    @classmethod
    def _replace_placeholder(cls, placeholder: str) -> str | None:
        """
        If the information or description field is empty, Nyaa replaces it with a placeholder value.
        This replaces said placeholder value with `None`.
        """
        if placeholder in ("No information.", "#### No description."):
            return None
        return placeholder

    def __eq__(self, other: object) -> bool:
        """
        Implements equality method.
        Two NyaaTorrentPage(s) are equal if they both point to the same URL.
        """
        if not isinstance(other, NyaaTorrentPage):
            return NotImplemented
        return self.url == other.url

    def __hash__(self) -> int:
        """
        Makes NyaaTorrentPage hashable.
        """
        return hash(self.url)

    def __repr__(self) -> str:
        """
        A shorter human readable __repr__ because
        the default one is too long.
        """
        body = f"title='{self.title}', url='{self.url}', category='{self.category}', submitter='{self.submitter}', datetime='{self.datetime.isoformat()}'"
        return f"{self.__class__.__name__}({body})"

    def __str__(self) -> str:
        """
        Stringify into something easily readable.
        """
        return self.title



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from httpx import AsyncClient, Client

from pynyaa import AsyncNyaa, Nyaa

if TYPE_CHECKING:
    from typing_extensions import AsyncGenerator, Generator

headers = {
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36"
}


@pytest.fixture
def nyaa_client() -> Generator[Nyaa]:
    with Nyaa(client=Client(headers=headers)) as nyaa:
        yield nyaa


@pytest.fixture
async def async_nyaa_client() -> AsyncGenerator[AsyncNyaa]:
    async with AsyncNyaa(client=AsyncClient(headers=headers)) as nyaa:
        yield nyaa



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from pynyaa._compat import IntEnum, StrEnum
from pynyaa._types import CategoryID, CategoryLiteral, SortByLiteral

if TYPE_CHECKING:
    from typing_extensions import Self

CATEGORY_NAME_TO_ID_MAP: dict[str, CategoryID] = {
    # All, c=0_0
    "All": "0_0",
    # Anime, c=1_X
    "Anime": "1_0",
    "Anime - Anime Music Video": "1_1",
    "Anime - English-translated": "1_2",
    "Anime - Non-English-translated": "1_3",
    "Anime - Raw": "1_4",
    # Audio, c=2_X
    "Audio": "2_0",
    "Audio - Lossless": "2_1",
    "Audio - Lossy": "2_2",
    # Literature, c=3_X
    "Literature": "3_0",
    "Literature - English-translated": "3_1",
    "Literature - Non-English-translated": "3_2",
    "Literature - Raw": "3_3",
    # Live Action, c=4_X
    "Live Action": "4_0",
    "Live Action - English-translated": "4_1",
    "Live Action - Idol/Promotional Video": "4_2",
    "Live Action - Non-English-translated": "4_3",
    "Live Action - Raw": "4_4",
    # Pictures, c=5_X
    "Pictures": "5_0",
    "Pictures - Graphics": "5_1",
    "Pictures - Photos": "5_2",
    # Software, c=6_X
    "Software": "6_0",
    "Software - Applications": "6_1",
    "Software - Games": "6_2",
}

CATEGORY_ID_TO_NAME_MAP = {v: k for k, v in CATEGORY_NAME_TO_ID_MAP.items()}


class BaseStrEnum(StrEnum):
    """StrEnum with case-insensitive double-sided lookup"""

    @classmethod
    def _missing_(cls, value: object) -> Self:
        errmsg = f"'{value}' is not a valid {cls.__name__}"

        if isinstance(value, str):
            for member in cls:
                if (member.value.casefold() == value.casefold()) or (member.name.casefold() == value.casefold()):
                    return member
            raise ValueError(errmsg)
        raise ValueError(errmsg)


class Category(BaseStrEnum):
    """Nyaa categories"""

    ALL = "All"

    ANIME = "Anime"
    ANIME_MUSIC_VIDEO = "Anime - Anime Music Video"
    ANIME_ENGLISH_TRANSLATED = "Anime - English-translated"
    ANIME_NON_ENGLISH_TRANSLATED = "Anime - Non-English-translated"
    ANIME_RAW = "Anime - Raw"

    AUDIO = "Audio"
    AUDIO_LOSSLESS = "Audio - Lossless"
    AUDIO_LOSSY = "Audio - Lossy"

    LITERATURE = "Literature"
    LITERATURE_ENGLISH_TRANSLATED = "Literature - English-translated"
    LITERATURE_NON_ENGLISH_TRANSLATED = "Literature - Non-English-translated"
    LITERATURE_RAW = "Literature - Raw"

    LIVE_ACTION = "Live Action"
    LIVE_ACTION_ENGLISH_TRANSLATED = "Live Action - English-translated"
    LIVE_ACTION_IDOL_PROMOTIONAL_VIDEO = "Live Action - Idol/Promotional Video"
    LIVE_ACTION_NON_ENGLISH_TRANSLATED = "Live Action - Non-English-translated"
    LIVE_ACTION_RAW = "Live Action - Raw"

    PICTURES = "Pictures"
    PICTURES_GRAPHICS = "Pictures - Graphics"
    PICTURES_PHOTOS = "Pictures - Photos"

    SOFTWARE = "Software"
    SOFTWARE_APPLICATIONS = "Software - Applications"
    SOFTWARE_GAMES = "Software - Games"

    @property
    def id(self) -> CategoryID:
        """
        Returns the ID of the category.

        Examples
        --------
        ```py
        >>> Category.ANIME_ENGLISH_TRANSLATED.id == "1_2"
        True
        >>> Category.ALL.id == "0_0"
        True
        ```
        """
        return CATEGORY_NAME_TO_ID_MAP[self.value]

    @property
    def parent(self) -> Self:
        """
        Returns the parent category.

        Examples
        --------
        ```py
        >>> Category.ANIME_ENGLISH_TRANSLATED.parent is Category.ANIME
        True
        >>> Category.ANIME.parent is Category.ANIME
        True
        ```
        """
        return self.get(self.value.split("-")[0].strip())

    @overload
    @classmethod
    def get(cls, key: CategoryLiteral, default: CategoryLiteral = "ALL") -> Self: ...

    @overload
    @classmethod
    def get(cls, key: CategoryLiteral, default: str = "ALL") -> Self: ...

    @overload
    @classmethod
    def get(cls, key: str, default: CategoryLiteral = "ALL") -> Self: ...

    @overload
    @classmethod
    def get(cls, key: str, default: str = "ALL") -> Self: ...

    @classmethod
    def get(cls, key: CategoryLiteral | str, default: CategoryLiteral | str = "ALL") -> Self:
        """
        Get the `Category` by its name, value, or id (case-insensitive).
        Return the default if the key is missing or invalid.

        Parameters
        ----------
        key : CategoryLiteral | str
            The key to retrieve.
        default : CategoryLiteral | str, optional
            The default value to return if the key is missing or invalid.

        Returns
        -------
        Category
            The `Category` corresponding to the key.
        """

        try:
            return cls(key)
        except ValueError:
            try:
                return cls(CATEGORY_ID_TO_NAME_MAP[key])  # type: ignore
            except KeyError:
                return cls(default)


class SortBy(BaseStrEnum):
    COMMENTS = "comments"
    SIZE = "size"
    DATETIME = "id"  # yea... https://nyaa.si/?s=id&o=desc
    SEEDERS = "seeders"
    LEECHERS = "leechers"
    DOWNLOADS = "downloads"

    @overload
    @classmethod
    def get(cls, key: SortByLiteral, default: SortByLiteral = "datetime") -> Self: ...

    @overload
    @classmethod
    def get(cls, key: SortByLiteral, default: str = "datetime") -> Self: ...

    @overload
    @classmethod
    def get(cls, key: str, default: SortByLiteral = "datetime") -> Self: ...

    @overload
    @classmethod
    def get(cls, key: str, default: str = "datetime") -> Self: ...

    @classmethod
    def get(cls, key: SortByLiteral | str, default: SortByLiteral | str = "datetime") -> Self:
        """
        Get the `SortBy` by its name or value (case-insensitive).
        Return the default if the key is missing or invalid.

        Parameters
        ----------
        key : SortByLiteral | str
            The key to retrieve.
        default : SortByLiteral | str, optional
            The default value to return if the key is missing or invalid.

        Returns
        -------
        Category
            The `SortBy` corresponding to the key.
        """
        try:
            return cls(key)
        except ValueError:
            return cls(default)


class Filter(IntEnum):
    """Nyaa search filters"""

    NO_FILTER = 0
    NO_REMAKES = 1
    TRUSTED_ONLY = 2



from __future__ import annotations

from datetime import datetime, timezone
from typing import Annotated, Literal, Union

from pydantic import AfterValidator, AnyUrl, UrlConstraints

UTCDateTime = Annotated[datetime, AfterValidator(lambda dt: dt.astimezone(timezone.utc))]
"""datetime.datetime that's always in UTC."""

MagnetUrl = Annotated[AnyUrl, UrlConstraints(allowed_schemes=["magnet"])]
"""Url that only allows magnets."""

CategoryName = Literal[
    "ALL",
    "ANIME",
    "ANIME_MUSIC_VIDEO",
    "ANIME_ENGLISH_TRANSLATED",
    "ANIME_NON_ENGLISH_TRANSLATED",
    "ANIME_RAW",
    "AUDIO",
    "AUDIO_LOSSLESS",
    "AUDIO_LOSSY",
    "LITERATURE",
    "LITERATURE_ENGLISH_TRANSLATED",
    "LITERATURE_NON_ENGLISH_TRANSLATED",
    "LITERATURE_RAW",
    "LIVE_ACTION",
    "LIVE_ACTION_ENGLISH_TRANSLATED",
    "LIVE_ACTION_IDOL_PROMOTIONAL_VIDEO",
    "LIVE_ACTION_NON_ENGLISH_TRANSLATED",
    "LIVE_ACTION_RAW",
    "PICTURES",
    "PICTURES_GRAPHICS",
    "PICTURES_PHOTOS",
    "SOFTWARE",
    "SOFTWARE_APPLICATIONS",
    "SOFTWARE_GAMES",
]

CategoryValue = Literal[
    "All",
    "Anime",
    "Anime - Anime Music Video",
    "Anime - English-translated",
    "Anime - Non-English-translated",
    "Anime - Raw",
    "Audio",
    "Audio - Lossless",
    "Audio - Lossy",
    "Literature",
    "Literature - English-translated",
    "Literature - Non-English-translated",
    "Literature - Raw",
    "Live Action",
    "Live Action - English-translated",
    "Live Action - Idol/Promotional Video",
    "Live Action - Non-English-translated",
    "Live Action - Raw",
    "Pictures",
    "Pictures - Graphics",
    "Pictures - Photos",
    "Software",
    "Software - Applications",
    "Software - Games",
]

CategoryID = Literal[
    "0_0",
    "1_0",
    "1_1",
    "1_2",
    "1_3",
    "1_4",
    "2_0",
    "2_1",
    "2_2",
    "3_0",
    "3_1",
    "3_2",
    "3_3",
    "4_0",
    "4_1",
    "4_2",
    "4_3",
    "4_4",
    "5_0",
    "5_1",
    "5_2",
    "6_0",
    "6_1",
    "6_2",
]

CategoryLiteral = Union[CategoryName, CategoryValue, CategoryID]

SortByLiteral = Literal["comments", "size", "id", "datetime", "seeders", "leechers", "downloads"]



from __future__ import annotations

import pytest

from pynyaa import Category, Filter, SortBy


def test_category_value_error() -> None:
    with pytest.raises(ValueError):
        Category.get("asdadadsad", "invalid default")


def test_sortby_value_error() -> None:
    with pytest.raises(ValueError):
        SortBy.get("asdadadsad", "invalid default")


@pytest.mark.parametrize(
    "category, expected_id",
    [
        (Category.ALL, "0_0"),
        (Category.ANIME, "1_0"),
        (Category.ANIME_MUSIC_VIDEO, "1_1"),
        (Category.ANIME_ENGLISH_TRANSLATED, "1_2"),
        (Category.ANIME_NON_ENGLISH_TRANSLATED, "1_3"),
        (Category.ANIME_RAW, "1_4"),
        (Category.AUDIO, "2_0"),
        (Category.AUDIO_LOSSLESS, "2_1"),
        (Category.AUDIO_LOSSY, "2_2"),
        (Category.LITERATURE, "3_0"),
        (Category.LITERATURE_ENGLISH_TRANSLATED, "3_1"),
        (Category.LITERATURE_NON_ENGLISH_TRANSLATED, "3_2"),
        (Category.LITERATURE_RAW, "3_3"),
        (Category.LIVE_ACTION, "4_0"),
        (Category.LIVE_ACTION_ENGLISH_TRANSLATED, "4_1"),
        (Category.LIVE_ACTION_IDOL_PROMOTIONAL_VIDEO, "4_2"),
        (Category.LIVE_ACTION_NON_ENGLISH_TRANSLATED, "4_3"),
        (Category.LIVE_ACTION_RAW, "4_4"),
        (Category.PICTURES, "5_0"),
        (Category.PICTURES_GRAPHICS, "5_1"),
        (Category.PICTURES_PHOTOS, "5_2"),
        (Category.SOFTWARE, "6_0"),
        (Category.SOFTWARE_APPLICATIONS, "6_1"),
        (Category.SOFTWARE_GAMES, "6_2"),
    ],
)
def test_nyaa_category_id_property(category, expected_id):
    assert category.id == expected_id


@pytest.mark.parametrize(
    "category, expected_parent",
    [
        (Category.ALL, Category.ALL),
        (Category.ANIME, Category.ANIME),
        (Category.ANIME_MUSIC_VIDEO, Category.ANIME),
        (Category.ANIME_ENGLISH_TRANSLATED, Category.ANIME),
        (Category.ANIME_NON_ENGLISH_TRANSLATED, Category.ANIME),
        (Category.ANIME_RAW, Category.ANIME),
        (Category.AUDIO, Category.AUDIO),
        (Category.AUDIO_LOSSLESS, Category.AUDIO),
        (Category.AUDIO_LOSSY, Category.AUDIO),
        (Category.LITERATURE, Category.LITERATURE),
        (Category.LITERATURE_ENGLISH_TRANSLATED, Category.LITERATURE),
        (Category.LITERATURE_NON_ENGLISH_TRANSLATED, Category.LITERATURE),
        (Category.LITERATURE_RAW, Category.LITERATURE),
        (Category.LIVE_ACTION, Category.LIVE_ACTION),
        (Category.LIVE_ACTION_ENGLISH_TRANSLATED, Category.LIVE_ACTION),
        (Category.LIVE_ACTION_IDOL_PROMOTIONAL_VIDEO, Category.LIVE_ACTION),
        (Category.LIVE_ACTION_NON_ENGLISH_TRANSLATED, Category.LIVE_ACTION),
        (Category.LIVE_ACTION_RAW, Category.LIVE_ACTION),
        (Category.PICTURES, Category.PICTURES),
        (Category.PICTURES_GRAPHICS, Category.PICTURES),
        (Category.PICTURES_PHOTOS, Category.PICTURES),
        (Category.SOFTWARE, Category.SOFTWARE),
        (Category.SOFTWARE_APPLICATIONS, Category.SOFTWARE),
        (Category.SOFTWARE_GAMES, Category.SOFTWARE),
    ],
)
def test_nyaa_category_parent_property(category, expected_parent):
    assert category.parent is expected_parent


@pytest.mark.parametrize(
    "input_value, expected_category",
    [
        ("All", Category.ALL),
        ("all", Category.ALL),
        ("Anime", Category.ANIME),
        ("Anime - Anime Music Video", Category.ANIME_MUSIC_VIDEO),
        ("Anime - English-translated", Category.ANIME_ENGLISH_TRANSLATED),
        ("Anime - Non-English-translated", Category.ANIME_NON_ENGLISH_TRANSLATED),
        ("Anime - Raw", Category.ANIME_RAW),
        ("Audio", Category.AUDIO),
        ("Audio - Lossless", Category.AUDIO_LOSSLESS),
        ("Audio - Lossy", Category.AUDIO_LOSSY),
        ("Literature", Category.LITERATURE),
        ("Literature - English-translated", Category.LITERATURE_ENGLISH_TRANSLATED),
        ("Literature - Non-English-translated", Category.LITERATURE_NON_ENGLISH_TRANSLATED),
        ("Literature - Raw", Category.LITERATURE_RAW),
        ("Live Action", Category.LIVE_ACTION),
        ("Live Action - English-translated", Category.LIVE_ACTION_ENGLISH_TRANSLATED),
        ("Live Action - Idol/Promotional Video", Category.LIVE_ACTION_IDOL_PROMOTIONAL_VIDEO),
        ("Live Action - Non-English-translated", Category.LIVE_ACTION_NON_ENGLISH_TRANSLATED),
        ("Live Action - Raw", Category.LIVE_ACTION_RAW),
        ("Pictures", Category.PICTURES),
        ("Pictures - Graphics", Category.PICTURES_GRAPHICS),
        ("Pictures - Photos", Category.PICTURES_PHOTOS),
        ("Software", Category.SOFTWARE),
        ("Software - Applications", Category.SOFTWARE_APPLICATIONS),
        ("Software - Games", Category.SOFTWARE_GAMES),
        ("Invalid Category", Category.ALL),
        ("aNiMe", Category.ANIME),
        ("aUdIo - lOsSy", Category.AUDIO_LOSSY),
        ("lIvE aCtIoN - eNgLiSh-TrAnSlAtEd", Category.LIVE_ACTION_ENGLISH_TRANSLATED),
    ],
)
def test_nyaa_category_get(input_value, expected_category):
    assert Category.get(input_value) == expected_category


@pytest.mark.parametrize(
    "key, category",
    [
        ("0_0", Category.ALL),
        ("1_0", Category.ANIME),
        ("1_1", Category.ANIME_MUSIC_VIDEO),
        ("1_2", Category.ANIME_ENGLISH_TRANSLATED),
        ("1_3", Category.ANIME_NON_ENGLISH_TRANSLATED),
        ("1_4", Category.ANIME_RAW),
        ("2_0", Category.AUDIO),
        ("2_1", Category.AUDIO_LOSSLESS),
        ("2_2", Category.AUDIO_LOSSY),
        ("3_0", Category.LITERATURE),
        ("3_1", Category.LITERATURE_ENGLISH_TRANSLATED),
        ("3_2", Category.LITERATURE_NON_ENGLISH_TRANSLATED),
        ("3_3", Category.LITERATURE_RAW),
        ("4_0", Category.LIVE_ACTION),
        ("4_1", Category.LIVE_ACTION_ENGLISH_TRANSLATED),
        ("4_2", Category.LIVE_ACTION_IDOL_PROMOTIONAL_VIDEO),
        ("4_3", Category.LIVE_ACTION_NON_ENGLISH_TRANSLATED),
        ("4_4", Category.LIVE_ACTION_RAW),
        ("5_0", Category.PICTURES),
        ("5_1", Category.PICTURES_GRAPHICS),
        ("5_2", Category.PICTURES_PHOTOS),
        ("6_0", Category.SOFTWARE),
        ("6_1", Category.SOFTWARE_APPLICATIONS),
        ("6_2", Category.SOFTWARE_GAMES),
        ("6_9", Category.ALL),
        ("9_9", Category.ALL),
        ("-1_-1", Category.ALL),
    ],
)
def test_nyaa_category_get_with_id(key, category):
    assert Category.get(key) is category


@pytest.mark.parametrize(
    "input_value, default, expected_category",
    [
        (12345, "Anime", Category.ANIME),
        (None, "Software", Category.SOFTWARE),
        (None, "aLL", Category.ALL),
        ("test", "lIvE aCtIoN - eNgLiSh-TrAnSlAtEd", Category.LIVE_ACTION_ENGLISH_TRANSLATED),
    ],
)
def test_nyaa_category_get_with_default(input_value, default, expected_category):
    assert Category.get(input_value, default=default) == expected_category


@pytest.mark.parametrize(
    "input_value, expected_sort_by",
    [
        ("comments", SortBy.COMMENTS),
        ("size", SortBy.SIZE),
        ("id", SortBy.DATETIME),
        ("datetime", SortBy.DATETIME),
        ("seeders", SortBy.SEEDERS),
        ("leechers", SortBy.LEECHERS),
        ("downloads", SortBy.DOWNLOADS),
        ("random", SortBy.DATETIME),
        ("", SortBy.DATETIME),
        ("COMMENTS", SortBy.COMMENTS),
        ("SeEdErS", SortBy.SEEDERS),
    ],
)
def test_sort_by_get(input_value, expected_sort_by):
    assert SortBy.get(input_value) == expected_sort_by


@pytest.mark.parametrize(
    "input_value, default, expected_sort_by",
    [
        ("invalid", "comments", SortBy.COMMENTS),
        (12345, "size", SortBy.SIZE),
        (None, "iD", SortBy.DATETIME),
        ("SnEEders", "dateTIME", SortBy.DATETIME),
    ],
)
def test_sort_by_get_with_default(input_value, default, expected_sort_by):
    assert SortBy.get(input_value, default=default) == expected_sort_by


def test_nyaa_filter_enum_values() -> None:
    assert Filter.NO_FILTER == 0
    assert Filter.NO_REMAKES == 1
    assert Filter.TRUSTED_ONLY == 2



from __future__ import annotations

import re
from typing import Any
from urllib.parse import urljoin

from bs4 import BeautifulSoup


def parse_nyaa_torrent_page(base_url: str, html: str) -> dict[str, Any]:
    """
    Parse the HTML page to extract relevant information

    Parameters
    ----------
    base_url : str
        Base URL used to construct the full URL from relative URLs.
    html : str
        HTML content of the Nyaa page.

    Returns
    -------
    dict[str, Any]
        A dictionary with the relevant information.
    """

    soup = BeautifulSoup(html, "lxml")

    if success := soup.find("div", class_="panel panel-success"):
        # Trusted uploads use `panel-success` for their green color
        body = success
        is_trusted = True
        is_remake = False
    elif remake := soup.find("div", class_="panel panel-danger"):
        # Remake uploads use `panel-danger` for their red color
        body = remake
        is_trusted = False
        is_remake = True
    else:
        # Default uploads use `panel-default`
        body = soup.find("div", class_="panel panel-default")  # type: ignore
        is_trusted = False
        is_remake = False

    title = soup.title.string.replace(":: Nyaa", "").strip()  # type: ignore

    rows = body.find("div", class_="panel-body").find_all("div", class_="row")  # type: ignore

    # ROW ONE
    row_one = rows[0].find_all("div", class_="col-md-5")
    category = row_one[0].get_text().strip()
    datetime = row_one[1]["data-timestamp"]

    # ROW TWO
    row_two = rows[1].find_all("div", class_="col-md-5")
    submitter = row_two[0].get_text().strip()
    if submitter.lower() != "anonymous":
        submitter_url = urljoin(base_url, row_two[0].find("a").get("href", f"/user/{submitter}"))
        submitter_status = row_two[0].find("a").get("title", None)

        if submitter_status is not None:
            if submitter_status.lower() == "trusted":
                submitter_trusted = True
                submitter_banned = False
            elif submitter_status.lower() == "banned user":
                submitter_trusted = False
                submitter_banned = True
            elif submitter_status.lower() == "banned trusted":
                submitter_trusted = True
                submitter_banned = True
            else:
                submitter_trusted = False
                submitter_banned = False
    else:
        submitter_url = base_url
        submitter_trusted = False
        submitter_banned = False

    seeders = row_two[1].get_text().strip()

    # ROW THREE
    row_three = rows[2].find_all("div", class_="col-md-5")
    information = row_three[0].get_text().strip()
    leechers = row_three[1].get_text().strip()

    # ROW FOUR
    row_four = rows[3].find_all("div", class_="col-md-5")
    completed = row_four[1].get_text().strip()

    # ROW FOOTER
    footer = body.find("div", class_="panel-footer clearfix").find_all("a")  # type: ignore
    torrent_file = urljoin(base_url, footer[0]["href"])
    magnet = footer[1]["href"]

    # DESCRIPTION
    description = soup.find(id="torrent-description").get_text().strip()  # type: ignore

    return dict(
        title=title,
        category=category,
        datetime=datetime,
        submitter=dict(
            name=submitter,
            url=submitter_url,
            is_trusted=submitter_trusted,
            is_banned=submitter_banned,
        ),
        information=information,
        seeders=seeders,
        leechers=leechers,
        completed=completed,
        is_trusted=is_trusted,
        is_remake=is_remake,
        description=description,
        torrent_file=torrent_file,
        magnet=magnet,
    )


def parse_nyaa_search_results(html: str, base_url: str = "https://nyaa.si") -> tuple[str, ...]:
    """
    Parses the HTML of a Nyaa search results page to extract torrent links.

    Parameters
    ----------
    html : str
        The HTML content of the Nyaa search results page.
    base_url : str, optional
        The base URL to use for constructing full links from relative links.

    Yields
    ------
    str
        The full URL of torrent page, in the order they were present.
    """
    parsed = re.findall(r"<a href=\"(/view/\d+)\" title=\".*\">.*</a>", html)
    return tuple(urljoin(base_url, relative_link) for relative_link in parsed)



from __future__ import annotations

from typing import TYPE_CHECKING

from httpx import HTTPStatusError

from pynyaa._clients._async import AsyncNyaa
from pynyaa._clients._sync import Nyaa
from pynyaa._enums import Category, Filter, SortBy
from pynyaa._models import NyaaTorrentPage, Submitter
from pynyaa._version import __version__, __version_tuple__

if TYPE_CHECKING:
    from typing_extensions import Generator


def get(page: int | str) -> NyaaTorrentPage:  # pragma: no cover
    """
    Shortcut for `pynyaa.Nyaa.get`.
    For more advanced or configurable usage, use the `pynyaa.Nyaa` client directly.
    """
    with Nyaa() as nyaa:
        return nyaa.get(page)


def search(
    query: str,
    *,
    category: Category = Category.ALL,
    filter: Filter = Filter.NO_FILTER,
    sort_by: SortBy = SortBy.DATETIME,
    reverse: bool = False,
) -> Generator[NyaaTorrentPage]:  # pragma: no cover
    """
    Shortcut for `pynyaa.Nyaa.search`.
    For more advanced or configurable usage, use the `pynyaa.Nyaa` client directly.
    """
    with Nyaa() as nyaa:
        yield from nyaa.search(query, category=category, filter=filter, sort_by=sort_by, reverse=reverse)


__all__ = (
    # Clients
    "AsyncNyaa",
    "Nyaa",
    # Top level API,
    "get",
    "search",
    # Enums
    "Category",
    "Filter",
    "SortBy",
    # Models
    "NyaaTorrentPage",
    "Submitter",
    # Exceptions
    "HTTPStatusError",
    # Version
    "__version__",
    "__version_tuple__",
)



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin

from httpx import AsyncClient
from torf import Torrent

from pynyaa._enums import Category, Filter, SortBy
from pynyaa._models import NyaaTorrentPage
from pynyaa._parser import parse_nyaa_search_results, parse_nyaa_torrent_page
from pynyaa._version import __version__

if TYPE_CHECKING:
    from typing_extensions import AsyncGenerator, Self


class AsyncNyaa:
    def __init__(self, base_url: str = "https://nyaa.si/", client: AsyncClient | None = None) -> None:
        """
        Async Nyaa client.

        Parameters
        ----------
        base_url : str, optional
            The base URL of Nyaa.
            This is used for constructing the full URL from relative URLs.
        client : Client, optional
            An [`httpx.Client`](https://www.python-httpx.org/api/#client) instance used to make requests to Nyaa.
        """
        self._base_url = base_url
        self._client = AsyncClient(headers={"user-agent": f"pynyaa/{__version__}"}) if client is None else client

    @property
    def base_url(self) -> str:
        """
        This is the base URL, used for constructing the full URL from relative URLs.
        """
        return self._base_url

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, *args: object) -> None:
        await self.close()

    async def close(self) -> None:
        """Close the connection"""
        await self._client.aclose()

    async def get(self, page: int | str) -> NyaaTorrentPage:
        """
        Retrieve information from a Nyaa torrent page.

        Parameters
        ----------
        page : int or str
            The torrent page.
            This can either be a URL like `https://nyaa.si/view/123456`
            or just the ID like `123456`

        Raises
        ------
        HTTPStatusError
            Nyaa returned a non 2xx response.

        Returns
        -------
        NyaaTorrentPage
            A NyaaTorrentPage object representing the retrieved data.
        """

        nyaa_id = page if isinstance(page, int) else page.split("/")[-1]
        nyaa_url = urljoin(self._base_url, f"/view/{nyaa_id}")

        nyaa = await self._client.get(nyaa_url)
        nyaa.raise_for_status()

        parsed = parse_nyaa_torrent_page(self._base_url, nyaa.text)

        # Get the torrent file and convert it to a torf.Torrent object
        response = await self._client.get(parsed["torrent_file"])
        response.raise_for_status()
        torrent = Torrent.read_stream(BytesIO(response.content))

        return NyaaTorrentPage(id=nyaa_id, url=nyaa_url, torrent=torrent, **parsed)  # type: ignore

    async def search(
        self,
        query: str,
        *,
        category: Category = Category.ALL,
        filter: Filter = Filter.NO_FILTER,
        sort_by: SortBy = SortBy.DATETIME,
        reverse: bool = False,
    ) -> AsyncGenerator[NyaaTorrentPage]:
        """
        Search for torrents on Nyaa.

        Parameters
        ----------
        query : str
            The search query.
        category : Category, optional
            The category to narrow down the search results.
        filter : Filter, optional
            Specifies the filter to apply to the search results.
        sort_by : SortBy, optional
            Defines how to sort the results.
        reverse : bool, optional
            Determines the order of the results: ascending if `True`, descending if `False`.

        Raises
        ------
        HTTPStatusError
            Nyaa returned a non 2xx response.

        Yields
        -------
        NyaaTorrentPage
            A NyaaTorrentPage object representing the retrieved data.
        """
        params: dict[str, Any] = dict(
            f=filter,
            c=category.id,
            q=query,
            s=sort_by,
            o="asc" if reverse else "desc",  # desc is the default on nyaa
        )

        nyaa = await self._client.get(self._base_url, params=params)
        nyaa.raise_for_status()
        results = parse_nyaa_search_results(nyaa.text)

        for link in results:
            yield await self.get(link)



# type: ignore
from __future__ import annotations

from datetime import datetime

from torf import Torrent

from pynyaa import Category, NyaaTorrentPage, Submitter


def test_submitter() -> None:
    a = Submitter(name="John", url="https://nyaa.si/user/john", is_trusted=True, is_banned=False)
    b = Submitter(name="John", url="https://nyaa.si/user/john", is_trusted=True, is_banned=False)
    c = Submitter(name="Jane", url="https://nyaa.si/user/jane", is_trusted=False, is_banned=False)

    assert a.__repr__() == "Submitter(name='John', url='https://nyaa.si/user/john', is_trusted=True, is_banned=False)"
    assert a.__str__() == "John"

    assert a == b
    assert a != c
    assert c != "other"
    assert set((a, b, c)) == {a, c} == {b, c}


def test_nyaa_torrent_page() -> None:
    submitter_a = Submitter(name="John", url="https://nyaa.si/user/john", is_trusted=True, is_banned=False)
    page_a = NyaaTorrentPage(
        id=123456,
        url="https://nyaa.si/view/123456",
        title="title",
        category=Category.ANIME_ENGLISH_TRANSLATED.value,
        submitter=submitter_a,
        datetime=datetime(year=2024, month=6, day=30),
        information=None,
        seeders=20,
        leechers=30,
        completed=100,
        is_trusted=False,
        is_remake=False,
        description=None,
        torrent_file="https://nyaa.si/download/123456.torrent",
        magnet="magnet:?xt=urn:btih:...&dn=...",
        torrent=Torrent.read(filepath="tests/__torrents__/ubuntu-22.04.4-live-server-amd64.iso.torrent"),
    )

    submitter_b = Submitter(name="John", url="https://nyaa.si/user/john", is_trusted=True, is_banned=False)
    page_b = NyaaTorrentPage(
        id=123456,
        url="https://nyaa.si/view/123456",
        title="title",
        category=Category.ANIME_ENGLISH_TRANSLATED.value,
        submitter=submitter_b,
        datetime=datetime(year=2024, month=6, day=30),
        information=None,
        seeders=34,
        leechers=23,
        completed=123,
        is_trusted=True,
        is_remake=False,
        description="description",
        torrent_file="https://nyaa.si/download/123456.torrent",
        magnet="magnet:?xt=urn:btih:...&dn=...",
        torrent=Torrent.read(filepath="tests/__torrents__/ubuntu-22.04.4-live-server-amd64.iso.torrent"),
    )

    submitter_c = Submitter(name="Jane", url="https://nyaa.si/user/jane", is_trusted=False, is_banned=False)
    page_c = NyaaTorrentPage(
        id=567890,
        url="https://nyaa.si/view/567890",
        title="title",
        category=Category.ANIME_ENGLISH_TRANSLATED.value,
        submitter=submitter_c,
        datetime=datetime(year=2024, month=6, day=30),
        information=None,
        seeders=34,
        leechers=23,
        completed=123,
        is_trusted=True,
        is_remake=False,
        description="description",
        torrent_file="https://nyaa.si/download/567890.torrent",
        magnet="magnet:?xt=urn:btih:...&dn=...",
        torrent=Torrent.read(filepath="tests/__torrents__/ubuntu-24.04-desktop-amd64.iso.torrent"),
    )

    assert (
        repr(page_a)
        == "NyaaTorrentPage(title='title', url='https://nyaa.si/view/123456', category='Anime - English-translated', submitter='John', datetime='2024-06-30T00:00:00+00:00')"
    )
    assert str(page_a) == "title"

    assert page_a == page_b
    assert page_a != page_c
    assert page_c != "other"
    assert set((page_a, page_b, page_c)) == {page_a, page_c} == {page_b, page_c}

    assert (
        page_a.model_dump_json()
        == """{"id":123456,"url":"https://nyaa.si/view/123456","title":"title","category":"Anime - English-translated","submitter":{"name":"John","url":"https://nyaa.si/user/john","is_trusted":true,"is_banned":false},"datetime":"2024-06-30T00:00:00Z","information":null,"seeders":20,"leechers":30,"completed":100,"is_trusted":false,"is_remake":false,"description":null,"torrent_file":"https://nyaa.si/download/123456.torrent","magnet":"magnet:?xt=urn:btih:...&dn=...","torrent":{"name":"ubuntu-22.04.4-live-server-amd64.iso","size":2104408064,"infohash":"4d29c6c02c97caad937d8a9b66b0bb1b6f7cbbfe","piece_size":262144,"private":null,"trackers":[["https://torrent.ubuntu.com/announce"],["https://ipv6.torrent.ubuntu.com/announce"]],"comment":"Ubuntu CD releases.ubuntu.com","creation_date":null,"created_by":"mktorrent 1.1","source":null,"files":[{"file":"ubuntu-22.04.4-live-server-amd64.iso","size":2104408064}]}}"""
    )
